package fr.mrdev023.vulkan_java.vk;

import static fr.mrdev023.vulkan_java.vk.VulkanError.vkCheck;

import java.nio.IntBuffer;
import java.nio.LongBuffer;
import java.util.ArrayList;
import java.util.List;

import org.lwjgl.system.MemoryStack;
import org.lwjgl.vulkan.KHRSurface;
import org.lwjgl.vulkan.KHRSwapchain;
import org.lwjgl.vulkan.VK10;
import org.lwjgl.vulkan.VkExtent2D;
import org.lwjgl.vulkan.VkImageViewCreateInfo;
import org.lwjgl.vulkan.VkSurfaceCapabilitiesKHR;
import org.lwjgl.vulkan.VkSurfaceFormatKHR;
import org.lwjgl.vulkan.VkSwapchainCreateInfoKHR;
import org.tinylog.Logger;

import fr.mrdev023.vulkan_java.window.Display;
import fr.mrdev023.vulkan_java.window.DisplayMode;

public class Swapchain {
    private final Device device;
    private final Surface surface;
    private final long vkSwapchain;
    private final List<Long> vkImages;
    private final List<Long> vkImageViews;
    private final VkSurfaceCapabilitiesKHR vkSurfaceCapabilities;
    private final VkSurfaceFormatKHR.Buffer vkSurfaceFormats;
    private final IntBuffer vkPresentModes;

    private final IntBuffer chosenImageCount;
    private final VkSurfaceFormatKHR chosenSurfaceFormat;
    private final int chosenPresentMode;
    private final VkExtent2D chosenExtent;

    public Swapchain(Device device, Surface surface) throws VulkanError {
        this.device = device;
        this.surface = surface;

        try (MemoryStack stack = MemoryStack.stackPush()) {
            vkSurfaceCapabilities = queryVkSurfaceCapabilities(stack);
            vkSurfaceFormats = queryVkPhysicalDeviceSurfaceFormats(stack);
            vkPresentModes = queryVkPhysicalDeviceSurfacePresentModes(stack);

            chosenSurfaceFormat = vkSurfaceFormats.get(0);
            chosenPresentMode = KHRSurface.VK_PRESENT_MODE_FIFO_KHR;
            chosenExtent = chooseSwapExtent(stack);
            chosenImageCount = stack.ints(vkSurfaceCapabilities.minImageCount() + 1);
            if (vkSurfaceCapabilities.maxImageCount() > 0
                    && chosenImageCount.get(0) > vkSurfaceCapabilities.maxImageCount()) {
                chosenImageCount.put(0, vkSurfaceCapabilities.maxImageCount());
            }

            VkSwapchainCreateInfoKHR createInfo = createVkSwapchainCreateInfo(stack, false);
            LongBuffer pSwapChain = stack.longs(VK10.VK_NULL_HANDLE);
            vkCheck(KHRSwapchain.vkCreateSwapchainKHR(device.getVkDevice(), createInfo, null, pSwapChain),
                    "Failed to create swapchain");

            vkSwapchain = pSwapChain.get(0);

            IntBuffer imageCount = stack.ints(0);
            KHRSwapchain.vkGetSwapchainImagesKHR(device.getVkDevice(), vkSwapchain, imageCount, null);

            LongBuffer pImages = stack.mallocLong(imageCount.get(0));
            KHRSwapchain.vkGetSwapchainImagesKHR(device.getVkDevice(), vkSwapchain, imageCount, pImages);

            vkImages = new ArrayList<>(imageCount.get(0));
            for (int i = 0; i < imageCount.get(0); i++) {
                vkImages.add(pImages.get(i));
            }

            vkImageViews = createVkImageViews(stack);
            Logger.debug("Swapchain created");
        }
    }

    public Device getDevice() {
        return device;
    }

    public Surface getSurface() {
        return surface;
    }

    public VkSurfaceCapabilitiesKHR getVkSurfaceCapabilities() {
        return vkSurfaceCapabilities;
    }

    public VkSurfaceFormatKHR.Buffer getVkSurfaceFormats() {
        return vkSurfaceFormats;
    }

    public IntBuffer getVkPresentModes() {
        return vkPresentModes;
    }

    public void destroy() {
        for (long imageView : vkImageViews) {
            VK10.vkDestroyImageView(device.getVkDevice(), imageView, null);
        }
        KHRSwapchain.vkDestroySwapchainKHR(device.getVkDevice(), vkSwapchain, null);
    }

    private VkSwapchainCreateInfoKHR createVkSwapchainCreateInfo(MemoryStack stack, boolean recreate) {
        VkSwapchainCreateInfoKHR createInfo = VkSwapchainCreateInfoKHR.calloc(stack);

        createInfo.sType(KHRSwapchain.VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR);
        createInfo.surface(surface.getVkSurface());

        // Image settings
        createInfo.minImageCount(chosenImageCount.get(0));
        createInfo.imageFormat(chosenSurfaceFormat.format());
        createInfo.imageColorSpace(chosenSurfaceFormat.colorSpace());
        createInfo.imageExtent(chosenExtent);
        createInfo.imageArrayLayers(1);
        createInfo.imageUsage(VK10.VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT);

        // Queue family indices
        createInfo.imageSharingMode(VK10.VK_SHARING_MODE_EXCLUSIVE); // If different queues between graphics and
                                                                     // present, use VK_SHARING_MODE_CONCURRENT
        createInfo.pQueueFamilyIndices(stack.ints(device.getGraphicsQueue().orElseThrow().getFamilyIndex()));

        createInfo.preTransform(vkSurfaceCapabilities.currentTransform());
        createInfo.compositeAlpha(KHRSurface.VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR);
        createInfo.presentMode(chosenPresentMode);
        createInfo.clipped(true);

        if (recreate) {
            createInfo.oldSwapchain(vkSwapchain);
        } else {
            createInfo.oldSwapchain(VK10.VK_NULL_HANDLE);
        }

        return createInfo;
    }

    private List<Long> createVkImageViews(MemoryStack stack) throws VulkanError {
        List<Long> vkImageViews = new ArrayList<>();

        for (long image : vkImages) {
            VkImageViewCreateInfo createInfo = createVkImageViewCreateInfo(stack, image);
            LongBuffer pImageView = stack.mallocLong(1);
            vkCheck(VK10.vkCreateImageView(device.getVkDevice(), createInfo, null, pImageView),
                    "Failed to create image views");
            vkImageViews.add(pImageView.get(0));
        }

        return vkImageViews;
    }

    private VkImageViewCreateInfo createVkImageViewCreateInfo(MemoryStack stack, long image) {
        VkImageViewCreateInfo createInfo = VkImageViewCreateInfo.calloc(stack);

        createInfo.sType(VK10.VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO);
        createInfo.image(image);
        createInfo.viewType(VK10.VK_IMAGE_VIEW_TYPE_2D);
        createInfo.format(chosenSurfaceFormat.format());

        createInfo.components().r(VK10.VK_COMPONENT_SWIZZLE_IDENTITY);
        createInfo.components().g(VK10.VK_COMPONENT_SWIZZLE_IDENTITY);
        createInfo.components().b(VK10.VK_COMPONENT_SWIZZLE_IDENTITY);
        createInfo.components().a(VK10.VK_COMPONENT_SWIZZLE_IDENTITY);

        createInfo.subresourceRange().aspectMask(VK10.VK_IMAGE_ASPECT_COLOR_BIT);
        createInfo.subresourceRange().baseMipLevel(0);
        createInfo.subresourceRange().levelCount(1);
        createInfo.subresourceRange().baseArrayLayer(0);
        createInfo.subresourceRange().layerCount(1);

        return createInfo;
    }

    private VkExtent2D chooseSwapExtent(MemoryStack stack) {
        DisplayMode displayMode = Display.getDisplayMode();

        VkExtent2D actualExtent = VkExtent2D.malloc(stack).set(displayMode.getWidth(), displayMode.getHeight());

        VkExtent2D minExtent = vkSurfaceCapabilities.minImageExtent();
        VkExtent2D maxExtent = vkSurfaceCapabilities.maxImageExtent();

        actualExtent.width(Math.clamp(actualExtent.width(), minExtent.width(), maxExtent.width()));
        actualExtent.height(Math.clamp(actualExtent.height(), minExtent.height(), maxExtent.height()));

        return actualExtent;
    }

    private VkSurfaceCapabilitiesKHR queryVkSurfaceCapabilities(MemoryStack stack) {
        VkSurfaceCapabilitiesKHR vkSurfaceCapabilities = VkSurfaceCapabilitiesKHR.malloc(stack);
        KHRSurface.vkGetPhysicalDeviceSurfaceCapabilitiesKHR(device.getPhysicalDevice().getVkPhysicalDevice(),
                surface.getVkSurface(),
                vkSurfaceCapabilities);

        return vkSurfaceCapabilities;
    }

    private VkSurfaceFormatKHR.Buffer queryVkPhysicalDeviceSurfaceFormats(MemoryStack stack) {
        VkSurfaceFormatKHR.Buffer vkSurfaceFormat;
        IntBuffer count = stack.ints(0);

        KHRSurface.vkGetPhysicalDeviceSurfaceFormatsKHR(device.getPhysicalDevice().getVkPhysicalDevice(),
                surface.getVkSurface(), count, null);
        if (count.get(0) == 0) {
            throw new RuntimeException("No surface formats found");
        }

        vkSurfaceFormat = VkSurfaceFormatKHR.malloc(count.get(0), stack);
        KHRSurface.vkGetPhysicalDeviceSurfaceFormatsKHR(device.getPhysicalDevice().getVkPhysicalDevice(),
                surface.getVkSurface(), count, vkSurfaceFormat);

        return vkSurfaceFormat;
    }

    private IntBuffer queryVkPhysicalDeviceSurfacePresentModes(MemoryStack stack) {
        IntBuffer vkPresentMode;
        IntBuffer count = stack.ints(0);

        KHRSurface.vkGetPhysicalDeviceSurfacePresentModesKHR(device.getPhysicalDevice().getVkPhysicalDevice(),
                surface.getVkSurface(), count, null);
        if (count.get(0) == 0) {
            throw new RuntimeException("No present modes found");
        }

        vkPresentMode = stack.mallocInt(count.get(0));
        KHRSurface.vkGetPhysicalDeviceSurfacePresentModesKHR(device.getPhysicalDevice().getVkPhysicalDevice(),
                surface.getVkSurface(), count, vkPresentMode);

        return vkPresentMode;
    }
}
